%{

/*
afe - audio file editor
Author: Klaus Wachter
e-mail: mail@klaus-wachter-1.de
Copyright: Freeware
Platform: LINUX
*/

#include <fcntl.h>
#include <termios.h>
#include <sys/soundcard.h>

#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif
#ifndef max
#define max(a,b) ((a) < (b) ? (b) : (a))
#endif

#define YY_READ_BUF_SIZE 1

struct termios terminal_flags_orig, terminal_flags;

int num;
char * string = NULL, * src = NULL, * dst = NULL;
int audio;
int sample_size = 8, sample_rate = 8000, stereo = 0;
char *buffer = NULL;
int buffer_size = 4096;
int frame_size;
int step_time = 50, step;
off_t src_offset, low, high, src_end, dst_offset, dst_end;
off_t *ptr;

%}

%%

	tcgetattr(0, &terminal_flags_orig);
	tcgetattr(0, &terminal_flags);
	(&terminal_flags)->c_lflag &= ~(ECHO|ICANON);
	tcsetattr(0, TCSANOW, &terminal_flags);

	#ifdef VERIFY_AUDIO
	set_audio(O_RDONLY); close(audio);
	#endif

	buffer = realloc(buffer, buffer_size);
	ptr = &dst_offset;
	set_frame_size();
	set_step();

	printf("Type \"?\" for help.\n");


\?	help();

[0-9]+	num = atoi(yytext); printf("num = %d\n", num);

\"[^\"]*\"	{
	string = realloc(string, yyleng - 1);
	yytext[yyleng - 1] = 0;
	strcpy(string, yytext + 1);
	printf("string = \"%s\"\n", string);
	}

src	set_src(); printf("src = \"%s\"\n", src);

dst	set_dst(); printf("dst = \"%s\"\n", dst);

sr	{
	sample_rate = num;
	set_audio(O_RDONLY); close(audio);
	set_step();
	printf("sample_rate = %d\n", sample_rate);
	}

ss	{
	if(sample_size == 8) sample_size = 16; else sample_size = 8;
	set_frame_size();
	set_step();
	printf("sample_size = %d\n", sample_size);
	}

st	{
	if(stereo == 0) stereo = 1; else stereo = 0;
	set_frame_size();
	set_step();
	printf("stereo = %d\n", stereo);
	}

t	{
	step_time = num; set_step();
	printf("step_time = %d\n", step_time);
	}

a	{low -= step;
	if(low < 0) low = 0;}

f	{low += step;
	if(low > high) low = high;}

j	{high -= step; 
	if(high < low) high = low;}

;	{high += step;
	if(high > src_end) high = src_end;}

l	{
	low = src_offset;
	printf("low = src_offset\n");
	}

L	{
	src_offset = low;
	printf("src_offset = low\n");
	}

h	{
	high = src_offset;
	printf("high = src_offset\n");
	}

H	{
	src_offset = high;
	printf("src_offset = high\n");
	}


so	ptr = &src_offset; printf("src_offset manipulation\n");

do	ptr = &dst_offset; printf("dst_offset manipulation\n");

=	*ptr = num * 1024; printf("offfset = %dk\n", num);

\<	*ptr -= step; if(*ptr < 0) *ptr = 0;

\>	*ptr += step;

\^	{
	*ptr = 0;
	printf("offset = <start>\n");
	}

\$	{
	if(ptr == &src_offset) *ptr = src_end;
	else *ptr = dst_end;
	printf("offset = <end>\n");
	}

o	(*ptr)++; printf("%s\n", (int)((*ptr) % 2) ? "odd" : "even");


p	src_offset = play_file(src, low, high);

c	{
	*ptr = play_file((ptr == &src_offset) ? src : dst,
	*ptr, (ptr == &src_offset) ? src_end : dst_end);
	}

R$	{
	dst_offset = record_file(dst, dst_offset);
	if(dst_offset > dst_end) dst_end = dst_offset;
	play_dst_last_second();
	}

C$	{
	copy_segment(); 
	if(dst_offset > dst_end) dst_end = dst_offset;
	play_dst_last_second();
	}

T$	truncate_dst();

\\	play_srcseg_first_second();

\/	play_srcseg_last_second();

P	play_dst_last_second();

q	quit(0);

\.	{
	printf("\
\n\
num = %d, string = \"%s\", src = \"%s\", dst = \"%s\",\n\
sample_rate = %d, sample_size = %d, stereo = %d, step_time = %d,\n\
low = %d, high = %d,\n\
ptr = %s, src_offset = %dk+%d, dst_offset = %dk+%d\n\
",
	num, string, src, dst,
	sample_rate, sample_size, stereo, step_time,
	low, high, 
	(ptr == &src_offset) ? "src" : "dst",
	(int)(src_offset / 1024), (int)(src_offset % 2),
	(int)(dst_offset / 1024), (int)(dst_offset % 2)
	);
	}

!.+$	printf("%s\n", yytext); system(yytext + 1);

%%

unblock (fd)
     int fd;
{
  int flag;
  flag = fcntl (fd, F_GETFL);
  fcntl (fd, F_SETFL, flag | O_NONBLOCK);
}

block (fd)
     int fd;
{
  int flag;
  flag = fcntl (fd, F_GETFL);
  fcntl (fd, F_SETFL, flag & ~(O_NONBLOCK));
}

set_audio (o_flags)
     int o_flags;
{
  if ((audio = open ("/dev/dsp", o_flags, 0)) == -1)
    {
      perror ("/dev/dsp");
      quit (1);
    }
  ioctl (audio, SNDCTL_DSP_SAMPLESIZE, &sample_size);
  ioctl (audio, SNDCTL_DSP_STEREO, &stereo);
  ioctl (audio, SNDCTL_DSP_SPEED, &sample_rate);
}

play_file (char *filename, off_t start, off_t end)
{
  off_t current, last_offset;
  int fd, n_read;
  char c;
  if ((fd = open (filename, O_RDONLY)) == -1)
    {
      perror (filename);
      return (start);
    }
  else
    {
      set_audio (O_WRONLY);
      lseek (fd, start, SEEK_SET);
      printf ("playing ... press <space> to stop");
      unblock (0);
      while (
	     ((c = getchar ()) == -1)
	     && ((current = lseek (fd, 0, SEEK_CUR)) < end)
	     && ((n_read = read (fd, buffer, buffer_size)) > 0))
	write (audio, buffer, min (n_read, (end - current)));
      last_offset = lseek (fd, 0, SEEK_CUR);
      block (0);
      unput (c);
      close (fd);
      close (audio);
      printf ("\n");
      return (last_offset);
    }
}

play_srcseg_first_second()
{
play_file(src, low, min(low + (frame_size * sample_rate), high));
}

play_srcseg_last_second ()
{
  off_t temp_offset;
      if ((temp_offset = max(high - (frame_size * sample_rate), low)) < 0)
	temp_offset = 0;
      play_file (src, temp_offset, high);
}

play_dst_last_second()
{
off_t temp_offset;
      if ((temp_offset = dst_offset - (frame_size * sample_rate)) < 0)
	temp_offset = 0;
      play_file (dst, temp_offset, dst_offset);
}

record_file (char *filename, off_t start)
{
  off_t last_offset;
  char c;
  int fd;
  if ((fd = open (filename, O_WRONLY | O_CREAT, 0666)) == -1)
    {
      perror (filename);
      return (start);
    }
  else
    {
      set_audio (O_RDONLY);
      printf ("recording ... press <space> to stop");
      unblock (0);
      lseek (fd, start, SEEK_SET);
      while ((c = getchar ()) == -1)
	{
	  read (audio, buffer, buffer_size);
	  if (write (fd, buffer, buffer_size) == -1)
	    {
	      perror (filename);
	      break;
	    }
	}
      last_offset = lseek (fd, 0, SEEK_CUR);
      block (0);
      unput (c);
      close (fd);
      close (audio);
      printf ("\n");
      return (last_offset);
    }
}

copy_segment ()
{
  int src_fd, dst_fd;
  off_t current;
  int n_read;
  if ((src_fd = open (src, O_RDONLY)) == -1)
    perror (src);
  if ((dst_fd = open (dst, O_WRONLY | O_CREAT, 0666)) == -1)
    perror (dst);
  if ((src_fd == -1) || (dst_fd == -1))
    {
      close (src_fd);
      close (dst_fd);
    }
  else
    {
      printf ("copying:");
      lseek (src_fd, low, SEEK_SET);
      lseek (dst_fd, min (dst_offset, lseek (dst_fd, 0, SEEK_END)), SEEK_SET);
      while (
	     ((current = lseek (src_fd, 0, SEEK_CUR)) < high)
	     && ((n_read = read (src_fd, buffer, buffer_size)) > 0))
	if (write (dst_fd, buffer, min (n_read, (high - current))) == -1)
	  {
	    perror (dst);
	    break;
	  }
      dst_offset = lseek (dst_fd, 0, SEEK_CUR);
      close (src_fd);
      close (dst_fd);
      printf ("\n");
    }
}

truncate_dst ()
{
  int fd;
  if ((fd = open (dst, O_WRONLY | O_CREAT, 0666)) == -1)
    perror (dst);
  else
    {
      ftruncate (fd, dst_offset);
      close (fd);
      printf ("dst truncated\n");
      dst_end = dst_offset;
    }
}

set_frame_size ()
{
  frame_size = 1;
  if (sample_size == 16)
    frame_size *= 2;
  if (stereo == 1)
    frame_size *= 2;
}

set_step ()
{
  step = frame_size * ((sample_rate * step_time) / 1000);
}

get_end (char *filename)
{
  int fd;
  off_t end;
  if ((fd = open (filename, O_RDONLY)) >= 0)
    end = lseek (fd, 0, SEEK_END);
  else
    end = 0;
  close (fd);
  return (end);
}

set_src ()
{
  src = realloc (src, strlen (string));
  strcpy (src, string);
  src_offset = low = 0;
  src_end = high = get_end (src);
}

set_dst ()
{
  dst = realloc (dst, strlen (string));
  strcpy (dst, string);
  dst_end = dst_offset = get_end (dst);
}

quit (int n)
{
  tcsetattr (0, TCSANOW, &terminal_flags_orig);
  printf ("quitting\n");
  exit (n);
}

help ()
{
  printf ("\
 ? -> display this help screen; q -> quit this program\n\
 <number> -> num = <number>; \"<string>\" -> string = <string>\n\
 src -> src = string, initialize; dst -> dst = string, initialize\n\
 sr -> sample_rate = num\n\
 st -> toggle between stereo and mono; ss -> toggle sample_size\n\
 t -> step_time = num\n\
 a -> decr low; f -> incr low; j -> decr high; ; -> incr high\n\
 l -> low = src_offset; L -> src_offset = low\n\
 h -> high = src_offset; H -> src_offset = high\n\
 so -> src_offset manipulation; do -> dst_offset manipulation\n\
 = -> offset = num kbyte\n\
 < -> decrement offset; > -> increment offset\n\
 ^ -> offset = <start>; $ -> offset = <end>\n\
 o -> toggle offset between odd and even\n\
 p -> play src segment [low, high], stop with <space>\n\
 c -> play file continuously, stop with <space>\n\
 \\ -> play first, / -> play last second of src segment [low, high]\n\
 P -> play last second of dst until dst_offset\n\
 R<enter> -> record dst, stop with <space>\n\
 C<enter> -> copy src segment [low, high] to dst at dst_offset\n\
 T<enter> -> truncate dst at dst_offset\n\
 . -> display settings\n\
 !<string><enter> -> execute the command <string>\n\
");
}
